require 'mime-types'
require 'webrick'

module Hyde
  # Branding and version
  VERSION = "0.1"
  attr_reader :VERSION
  VLINE = "<ADDRESS>\n      Hyde/#{Hyde::VERSION} on WEBrick/#{WEBrick::VERSION} (Ruby/#{RUBY_VERSION}/#{RUBY_RELEASE_DATE})\n    </ADDRESS>"
  attr_reader :VLINE
  
  class Server < WEBrick::HTTPServer
    def initialize(config={},&setup)
      super(config)
      @hyde_pathspec = Hyde::Pathspec.new "/", &setup
      self.mount_proc '/' do |req,res|
        @hyde_pathspec.match(Hyde::Context.new(req.path, req, res))
      end
    end
  end

  module ErrorPages
    # 404 text
    def error404(request, filepath)
      request.response.status = 404
      if request.handles.include? 404
        request.response.body = request.handles[404].call(
          filepath
        ) 
      else
        request.response.body = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\">\n<HTML>\n  <HEAD><TITLE>File not found</TITLE></HEAD>\n  <BODY>\n    <H1>File not found</H1>\n    #{filepath}\n    <HR>\n    #{Hyde::VLINE}\n  </BODY>\n</HTML>"      
      end
      request.response["Content-Type"] = 'text/html'
    end
    module_function :error404
  end
  
  # Interchangeable glob/regex/string pattern matching
  module PatternMatching
    def prep_path(path,match_full_path: false)
      raise Exception, "Not permitted" if @lock_methods
      @match_full_path = match_full_path
      @path = normalize(path) if path.kind_of? String
      @path = path if path.kind_of? Regexp
    end
    def match?(path)
      raise Exception, "Not permitted" if @lock_methods
      # behvaiour used by "index" method
      return true if @path == ""
      split_path = path.split("/").filter { |x| x != "" }
      if @path.kind_of? Regexp then
        # this chunk of fuck is needed to force regexp into 3 rules:
        # 1) unsafe regexp means match the whole (remaining) line.
        # 3) safe regexp means match only the part before the next slash
        # 2) a .match? only returns when there are no leftovers
        # this forces the matching to work somewhat consistently
        test = @path.match normalize_input(path) if @match_full_path
        test = @path.match split_path[0] unless @match_full_path
        if test and (test.pre_match == "") and (test.post_match == "") then
          return true
        else
          return false
        end
      else
        # algorithm to match path segments until no more left in @path
        @path.split("/").filter { |x| x != "" }
        .each_with_index { |x,i| 
          return false if x != split_path[i]
        }
        return true
      end
    end
    def normalize_input(path)
      raise Exception, "Not permitted" if @lock_methods
      # remove duplicate slashes and trim edge slashes
      (path.split "/").filter { |x| x != "" }.join("/")
    end
    def normalize(path)
      raise Exception, "Not permitted" if @lock_methods
      # remove duplicate slashes and trim edge slashes
      path = normalize_input(path)
      # globbing behaviour simulated with regexp
      if path.match /(?<!\\)[\*\?\[]/ then
        path = Regexp.new(path
          .gsub(/[\^\$\.\|\+\(\)\{\}]/,"\\\\\\0")
          .gsub(/(?<!\\)\*\*/,".*")
          .gsub(/(?<![.\\])\*/,"[^/]*")
          .gsub(/(?<!\\)\?/,".")
          .gsub(/(?<!\\)\[\!/,"[^")
        )
      end
      return path
    end
  end

  # Methods to control requests, accessible from within blocks
  module PublicContextControlMethods
    def redirect(url)
      puts "Unimplemented method 'redirect' called"
      return
    end
  end

  # Request wrapper class
  class Context
    def initialize(path,request,response)
      @path = path
      @filepath = ""
      @request = request
      @response = response
      @handles = {}
      @indexlist = []
      @vars = {}
    end
    attr_reader :request
    attr_reader :response
    attr_accessor :filepath
    attr_accessor :path
    attr_accessor :handles
    attr_accessor :indexlist
    attr_accessor :vars
  end

  # Context object with safe path encapsulation
  class ProtectedContext < Context
    def initialize(request)
      @path = request.path
      @filepath = request.filepath
      @request = request.request
      @response = request.response
      @handles = request.handles
      @indexlist = request.indexlist
      @vars = request.vars
    end
    undef :path=
    undef :filepath=
    undef :handles=
    undef :indexlist=
  end

  # Handler classes
  class Probe
    include Hyde::PatternMatching
    include Hyde::PublicContextControlMethods
    def initialize (path, match_full_path: false, &block_optional) 
      prep_path path, match_full_path: match_full_path
      @block = block_optional 
    end
    def match(request)
      if @block and (match? request.path) then
        @current_request = Hyde::ProtectedContext.new(request)
        @lock_methods = true
        return_later = self.instance_exec @current_request, &@block
        @lock_methods = false
        return return_later
      end
    end
  end

  class Serve < Hyde::Probe
    def match(request)
      return super if @block
      if match? request.path then
        match_path = normalize_input(request.path).match(@path)[0]
        filepath = request.filepath+match_path
        begin
          mimetype = MIME::Types.type_for(filepath)
          file = File.new filepath, "r"
          data = file.read()
          request.response.body = data
          request.response["Content-Type"] = mimetype
        rescue Errno::ENOENT
          Hyde::ErrorPages::error404 request, request.request.path
        end
      end
    end
  end

  class PrintProbe < Hyde::Probe
    def match(request)
      if match? request.path then
        puts "#{request.path} matched!"
      end
    end
  end 

  # Handler invocation functions
  module Handlers
    { 
      probe: Hyde::Probe,
      printProbe: Hyde::PrintProbe,
      serve: Hyde::Serve
    }.each_pair { |name, newclass|
      define_method name do |path, *a, **b, &block| 
        if path.kind_of? Array then
          path.each do |x|
            @chain.append newclass.new x, *a, **b, &block
          end
        else
          @chain.append newclass.new path, *a, **b, &block
        end
      end
    }
  end
  
  class Pathspec
    include Hyde::PatternMatching
    include Hyde::Handlers
    def initialize (path, root_path: nil, match_full_path: false, &block)
      prep_path path, match_full_path: match_full_path
      @chain = []
      @handles = {}
      @root_override = root_path
      @remap = false
      @lock_methods = true
      self.instance_exec &block
      @lock_methods = false
    end
    def path(path, *a, **b, &block)
      if path.kind_of? Array then
        path.each do |x|
          @chain.append Hyde::Pathspec.new x, *a, **b, &block
        end
      else
        @chain.append Hyde::Pathspec.new path, *a, **b, &block
      end
    end
    def root(path)
      @lock_methods = false
      @root_override = "/"+normalize_input(path)
      @lock_methods = true
    end
    def remap(path)
      @lock_methods = false
      @root_override = "/"+normalize_input(path)
      @lock_methods = true
      @remap = true
    end
    def index(list)
      @indexlist = list if list.kind_of? Array
    end
    def handle(code, &block)
      @handles[code] = block
    end
    def preprocess(&block) 
      @preprocessor = block
    end
    def match(request) 
      raise Exception, "Not permitted" if @lock_methods
      self.instance_exec request, &@preprocessor if @preprocessor
      if match? request.path then
        match_path = normalize_input(request.path).match(@path)
        next_path = match_path[0]
        request.path = cut_path = match_path.post_match
        # remap/root method handling
        if @root_override then
          request.filepath = if @remap then
                               @root_override+"/" 
                             else @root_override+"/"+next_path+"/" end
        else
          request.filepath = request.filepath+next_path+"/"
        end
        # parameter overrides overlaying
        @handles.each_pair { |k,v| request.handles[k] = v }
        request.indexlist = @indexlist if @indexlist
        # do directory indexing
        if cut_path.match /^\/?$/ then
          request.indexlist.each do |x| 
            try_index = @chain.find { |y| y.match? x }
            if try_index then
              request.path = x
              return try_index.match request
            end
          end
        end
        # passthrough to the next path object
        next_pathspec = @chain.find { |x| x.match? cut_path }
        next_pathspec.match request if next_pathspec
        unless next_pathspec then
          # throw 404 if nowhere to go
          Hyde::ErrorPages::error404 request, request.request.path
        end
      end
    end
  end
end
